from __future__ import annotations

DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.3.1 Safari/605.1.15"
)

MAX_INSTALL_RETRIES = 3

HTTP_REQUEST_TIMEOUT_SECONDS = 30
HTTP_STREAM_CONNECT_TIMEOUT_SECONDS = 10
HTTP_STREAM_READ_TIMEOUT_SECONDS = 120

HTTP_RETRY_TOTAL = 3
HTTP_RETRY_BACKOFF_FACTOR = 1
HTTP_RETRY_STATUS_FORCELIST = [429, 500, 502, 503, 504]
HTTP_RETRY_ALLOWED_METHODS = ["HEAD", "GET", "OPTIONS", "POST"]
